import logging
import os
import pathlib
import shutil
from mimetypes import guess_type

from dq.logging import error

logger = logging.getLogger(__name__)

# Incomplete list of textual types. These types are suitable for gzip.
TEXT_TYPES = set([
    'application/ecmascript',
    'application/javascript',
    'application/json',
    'application/x-javascript',
    'application/xhtml+xml',
    'application/xml',
    'image/svg+xml',
    'text/css',
    'text/csv',
    'text/html',
    'text/javascript',
    'text/markdown',
    'text/markdown',
    'text/plain',
    'text/xml',
])


def mkdirp(path):
    """Safely mkdir, creating all parent folders if they don't yet exist.

    This doesn't raise an error if the folder already exists. However, it does
    raise ``FileExistsError`` if the path points to an existing file.

    :param string path: The path to the folder to be created.
    """
    pathlib.Path(path).mkdir(parents=True, exist_ok=True)


def rmrf(path):
    """Remove a path like rm -rf.

    :param string path: The path to remove.
    """
    try:
        shutil.rmtree(path)
    except NotADirectoryError:
        os.remove(path)
    except FileNotFoundError:
        return


def traverse(path, callback):
    """Traverse a directory recursively, performing a task.

    :param string path: The path of the directory.
    :param func callback: A callback applied to each child file and directory.
        It takes 2 arguments, the relative path from the root, and whether the
        item is a directory.
    """
    if not os.path.isdir(path):
        raise (
            NotADirectoryError(20, 'Not a directory', path)
            if os.path.exists(path) else
            FileNotFoundError(2, 'No such file or directory', path)
        )

    path = os.path.join(path, '')
    pathlen = len(path)
    for cwd, dirs, files in os.walk(path):
        for d in dirs:
            callback(os.path.join(cwd, d)[pathlen:], True)
        for f in files:
            callback(os.path.join(cwd, f)[pathlen:], False)


def fileinfo(path):
    """Information of a file.

    This function throws an error if the path is not a valid file. To suppress
    errors, use ``safe_fileinfo`` instead.

    :param string path: The path to the file.
    :returns string mime: The MIME type of the file, such as text/plain.
    :returns int size: The size (in bytes) of the file.
    :returns string encoding: The encoding of the file, like gzip. This is
        suitable for use as the Content-Encoding header.
    """
    mime, encoding = guess_type(path)
    size = os.path.getsize(path)
    return mime, size, encoding


def safe_fileinfo(path):
    """Retrive information of a file without throwing an error.

    If the path is invalid, None, 0, None will be returned.

    :param string path: The path to the file.
    :returns string mime: The MIME type of the file, such as text/plain.
    :returns int size: The size (in bytes) of the file.
    :returns string encoding: The encoding of the file, like gzip. This is
        suitable for use as the Content-Encoding header.
    """
    try:
        return fileinfo(path)
    except Exception as e:
        error(logger, 'Error getting fileinfo', {'path': path, 'error': e})
        return None, 0, None



import logging
from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.sql.expression import Insert
from sqlalchemy.util import safe_reraise

from dq.config import Config
from dq.errors import ModelError
from dq.funcutil import safe_invoke
from dq.logging import error

logger = logging.getLogger(__name__)

engine = create_engine(
    Config.get('sql.url'),
    pool_recycle=600,
    pool_pre_ping=True,
    isolation_level='READ COMMITTED',
)
if Config.get('sql.flavor') == 'mysql':
    Insert.argument_for('mysql', 'replace_insert', None)


def session_maker():
    """Generate a scoped session. In general, use the Session global variable.

    :returns Session: A new SQLAlchemy session.
    """
    return scoped_session(sessionmaker(bind=engine))


Session = session_maker()


@compiles(Insert, 'mysql')
def replace_insert(insert, compiler, **kw):
    """Allow replace into for insert command.

    This only works for MySQL. It's not standard SQL. To enable this function,
    the sql.flavor config must be set to mysql.
    """
    s = compiler.visit_insert(insert, **kw)
    if 'mysql_replace_insert' in insert.kwargs:
        s = s.replace('INSERT INTO', 'REPLACE INTO')
    return s


@contextmanager
def commit_scope(session=None):
    """Commit any database operations within this scope."""
    session = session or Session()
    try:
        yield session
        session.commit()
    except Exception:
        with safe_reraise():
            session.rollback()


def unsafe_save_to_database(model, session=None):
    """Convenient method for writing things to the database.

    If the object is a list, its elements will be written to the database one
    by one. If the object is already in the database (i.e. when we are updating
    existing records), the caller of this function is responsible of ensuring
    that the same session used when retrieving the records is passed in. If
    a different session is provided, this operation will fail. (This is
    usually not an issue for view functions, which have session consistency on
    a thread level.)

    :param object/list model: The model object(s) to save. If a list is
        provided, the elements will be added one by one. Otherwise the object
        is added as-is.
    :param Session session: The optional database session to use.
    """
    if not model:
        return
    with commit_scope(session=session) as sess:
        if isinstance(model, list):
            for m in model:
                sess.add(m)
        else:
            sess.add(model)


def save_to_database(model, session=None):
    """Write things to the database.

    This function simply calls ``unsafe_save_to_database`` under the hood, but
    wraps the error into a ``ModelError`` upon failure.

    :param object/list model: The model object(s) to save. If a list is
        provided, the elements will be added one by one. Otherwise the object
        is added as-is.
    :param Session session: The optional database session to use.
    :raises ModelError: if the save fails.
    """
    try:
        unsafe_save_to_database(model, session=session)
    except Exception as e:
        error(logger, 'Error saving to database', {
            'model': model, 'error': e,
        })
        raise ModelError(e.message if hasattr(e, 'message') else str(e))


@safe_invoke(noreturn=True)
def safe_save_to_database(model, session=None):
    """Safely write things to the database.

    This function simply calls ``unsafe_save_to_database`` under the hood, but
    returns whether the operation is successful instead of throwing an error
    upon failure.

    :param object/list model: The model object(s) to save. If a list is
        provided, the elements will be added one by one. Otherwise the object
        is added as-is.
    :param Session session: The optional database session to use.
    :returns boolean: True if the operation is successful.
    """
    unsafe_save_to_database(model, session=session)


def query_with_limit_offset(query, limit, offset):
    """Add limit and offset constraints to a SQLAlchemy query.

    :param Query query: A SQLAlchemy query.
    :param int limit: The limit to add. If None or 0, it will be skipped.
    :param int offset: The offset to add. If None or 0, it will be skipped.
    :returns Query: The updated query object.
    """
    if limit:
        query = query.limit(limit)
    if offset:
        query = query.offset(offset)
    return query



import logging

import emails

from dq.config import Config
from dq.logging import error

logger = logging.getLogger(__name__)

SMTP_CONFIG = Config.get('mail')


def send_email(to, subject, html, sender_name, sender_email):
    """Send an email.

    :param string to: The recepient of the email.
    :param string subject: The subject of the email.
    :param string html: The HTML content of the email.
    :param string sender_name: The name of the sender.
    :param string sender_email: The sender address of the email.
    :returns boolean: True if the email is sent successfully. False otherwise.
    """
    mail = emails.Message(
        html=html,
        subject=subject,
        mail_from=(sender_name, sender_email),
    )
    resp = mail.send(to=to, smtp=SMTP_CONFIG)
    if not resp.success:
        error(logger, 'Failed to send email', {
            'error': resp.error.strerror, 'to': to, 'subject': subject,
        })
    return resp.success



import threading
import unittest
from time import sleep
from uuid import uuid4

from dq.redis import Redis


class TestRedis(unittest.TestCase):

    def test_exists(self):
        key = 'dqtest-{}'.format(uuid4())
        assert not Redis.exists(key)
        Redis.set(key, '1')
        assert Redis.exists(key)
        assert Redis.delete(key)
        assert not Redis.exists(key)

    def test_get(self):
        key = 'dqtest-{}'.format(uuid4())
        assert not Redis.get(key)
        Redis.setex(key, '1', 1)
        assert Redis.get(key) == '1'
        assert Redis.delete(key)
        assert not Redis.get(key)

    def test_get_json(self):
        key = 'dqtest-{}'.format(uuid4())
        assert not Redis.get(key)
        Redis.setex(key, [1, '2', True], 1)
        assert Redis.get(key) == '[1, "2", true]'
        Redis.setex(key, {'cornell': '#1'}, 1)
        assert Redis.get(key) == '{"cornell": "#1"}'

    def test_expire(self):
        key = 'dqtest-{}'.format(uuid4())
        assert Redis.set(key, '1')
        assert Redis.expire(key, 1)
        sleep(1.5)
        assert not Redis.exists(key)

    def test_hash(self):
        key = 'dqtest-{}'.format(uuid4())
        assert not Redis.hgetall(key)
        assert not Redis.hget(key, '1')
        assert Redis.hset(key, '1', '1')
        assert Redis.hget(key, '1') == '1'
        assert Redis.hset(key, '2', '3')
        assert Redis.hget(key, '2') == '3'
        assert Redis.hset(key, '3', {'cornell': '#1'})
        assert Redis.hget(key, '3') == '{"cornell": "#1"}'
        assert Redis.hgetall(key) == {
            '1': '1', '2': '3', '3': '{"cornell": "#1"}'
        }
        assert Redis.hdelete(key, '1')
        assert Redis.hgetall(key) == {'2': '3', '3': '{"cornell": "#1"}'}
        assert Redis.delete(key)
        assert not Redis.exists(key)

    def test_list(self):
        key = 'dqtest-{}'.format(uuid4())
        assert not Redis.lpeek(key, 1)
        assert not Redis.lpop(key, 1)
        assert Redis.rpush(key, 1, 2)
        assert Redis.rpush(key, '3', [1, 2])
        assert Redis.lpeek(key, 1) == ['1']
        assert Redis.lpeek(key, 2) == ['1', '2']
        assert Redis.lpop(key, 2) == ['1', '2']
        assert Redis.lpop(key, 3) == ['3', '[1, 2]']
        assert not Redis.exists(key)

    def test_atomic_rw(self):
        def evaluator(x):
            return x + '#1'

        Redis.set('dqtest-rw', 'cornell ')
        success, no_error = Redis.atomic_rw('dqtest-rw', evaluator)
        assert success and no_error
        assert Redis.get('dqtest-rw') == 'cornell #1'
        Redis.delete('dqtest-rw')

    def test_atomic_rw_user_abort(self):
        def evaluator(x):
            return None

        Redis.set('dqtest-rw', 'meow')
        success, no_error = Redis.atomic_rw('dqtest-rw', evaluator)
        assert not success and no_error
        assert Redis.get('dqtest-rw') == 'meow'
        Redis.delete('dqtest-rw')

    def test_atomic_rw_error(self):
        def change():
            Redis.set('dqtest-rw', 'cornell')

        def evaluator(x):
            t = threading.Thread(target=change)
            t.start()
            sleep(0.5)
            return x + '#1'

        Redis.set('dqtest-rw', 'meow')
        success, no_error = Redis.atomic_rw('dqtest-rw', evaluator)
        assert not success and not no_error
        assert Redis.get('dqtest-rw') == 'cornell'
        Redis.delete('dqtest-rw')

    def test_atomic_rw_hash(self):
        def evaluator(x):
            return x + '#1'

        Redis.hset('dqtest-rwh', 'key', 'cornell ')
        succ, no_error = Redis.atomic_rw_hash('dqtest-rwh', 'key', evaluator)
        assert succ and no_error
        assert Redis.hget('dqtest-rwh', 'key') == 'cornell #1'
        Redis.delete('dqtest-rwh')

    def test_atomic_rw_hash_user_abort(self):
        def evaluator(x):
            return None

        Redis.hset('dqtest-rwh', 'key', 'meow')
        succ, no_error = Redis.atomic_rw_hash('dqtest-rwh', 'key', evaluator)
        assert not succ and no_error
        assert Redis.hget('dqtest-rwh', 'key') == 'meow'
        Redis.delete('dqtest-rwh')

    def test_atomic_rw_hash_error(self):
        def change():
            Redis.hset('dqtest-rwh', 'key', 'meow2')

        def evaluator(x):
            t = threading.Thread(target=change)
            t.start()
            sleep(0.5)
            return x + '#1'

        Redis.hset('dqtest-rwh', 'key', 'meow')
        succ, no_error = Redis.atomic_rw_hash('dqtest-rwh', 'key', evaluator)
        assert not succ and not no_error
        assert Redis.hget('dqtest-rwh', 'key') == 'meow2'
        Redis.delete('dqtest-rwh')



import mock
import unittest

from dq import email


class TestEmail(unittest.TestCase):

    @mock.patch('emails.Message.send')
    def test_send_email(self, mock_send):
        resp = mock.MagicMock()
        resp.success = True
        mock_send.return_value = resp
        assert email.send_email(
            'unittest.to@danqing.io',
            'Unit test',
            'This is a test',
            'Danqing Liu',
            'unittest.from@danqing.io',
        )
        mock_send.assert_called_with(
            to='unittest.to@danqing.io',
            smtp={'user': 'danqing'},
        )

    @mock.patch('dq.email.error')
    @mock.patch('emails.Message.send')
    def test_send_email_error(self, mock_send, mock_error):
        resp_error = mock.MagicMock()
        resp_error.strerror = 'not found'
        resp = mock.MagicMock()
        resp.error = resp_error
        resp.success = False
        mock_send.return_value = resp
        assert not email.send_email(
            'unittest.to@danqing.io',
            'Unit test',
            'This is a test',
            'Danqing Liu',
            'unittest.from@danqing.io',
        )
        mock_send.assert_called_with(
            to='unittest.to@danqing.io',
            smtp={'user': 'danqing'},
        )
        mock_error.assert_called()



import os
import unittest

from dq.config import Config


class TestConfig(unittest.TestCase):

    def test_get(self):
        assert Config.get('author.email') == 'unittest@danqing.io'
        assert not Config.get('author.phone')
        assert os.getenv('NUMBER1') == 'cornell'

    def test_get_local2(self):
        os.environ['DQENV'] = 'local2'
        os.environ['NUMBER1'] = ''
        Config._instance = None
        assert Config.get('author.email') == 'unittest2@danqing.io'
        assert not os.getenv('NUMBER1')
        os.environ['DQENV'] = ''
        Config._instance = None



import unittest

from dq import errors


class TestErrors(unittest.TestCase):

    def test_recoverable_error(self):
        e = errors.RecoverableError('recoverable')
        assert e.status_code == 500
        assert e.message == 'recoverable'
        assert e.to_dict() == {'error': 'DQ Error'}
        assert e.__str__() == '500 - recoverable'
        assert e.__repr__() == (
            '<RecoverableError - status: 500, message: recoverable>')

    def test_model_error(self):
        e = errors.ModelError('model')
        assert e.status_code == 500
        assert e.message == 'model'

    def test_integration_error(self):
        e = errors.IntegrationError(None)
        assert e.status_code == 500
        assert not e.message
        assert e.__str__() == '500 - Unspecified DQ Error'

    def test_invalid_input_error(self):
        e = errors.InvalidInputError('invalid')
        assert e.status_code == 400
        assert e.to_dict() == {'error': 'invalid'}
        assert e.__str__() == '400 - invalid'

    def test_not_found_error(self):
        e = errors.NotFoundError()
        assert e.status_code == 404
        assert e.message == 'Not found'

    def test_unauthorized_error(self):
        e = errors.UnauthorizedError()
        assert e.status_code == 401
        assert e.message == 'Unauthorized'

    def test_forbidden_error(self):
        e = errors.ForbiddenError()
        assert e.status_code == 403
        assert e.message == 'Forbidden'



import json

import redis

from dq.config import Config


def strval(value):
    """JSON serialize value as appropriate.

    This function should only be used internally.

    :param dict|list|string|number value: An input value.
    :returns string: The output value, suitable for saving by Redis. If
        ``value`` is a ``dict`` or ``list``, it will be JSON-serialized.
        Otherwise it will be left as-is. Note that while Redis only takes
        string values, numbers have their string values be themselves in
        strings, and the conversion will be done by Redis automatically.
    """
    return json.dumps(value) if isinstance(value, (list, dict)) else value


def strvals(*values):
    """JSON serialize values as appropriate.

    This function should only be used internally.

    :param ...dict|list|string|number values: Input values.
    :returns list<string>: The output values. See docs for ``strval`` for
        more explanations.
    """
    return [strval(v) for v in values]


class Redis(object):

    _instance = redis.StrictRedis(**Config.get('redis'))

    @classmethod
    def exists(cls, key):
        """Whether the key exists in Redis.

        :param string key: The Redis key.
        :returns boolean: ``True`` if the key exists, and ``False`` otherwise.
        """
        return cls._instance.exists(key)

    @classmethod
    def get(cls, key):
        """Get the value stored at the key.

        :param string key: The Redis key.
        :returns string: The value of the key. If the key does not exist,
            ``None`` will be returned.
        """
        return cls._instance.get(key)

    @classmethod
    def set(cls, key, value):
        """Set the key to the specified value.

        :param string key: The Redis key.
        :param string value: The value to set. If this is not a string, it will
            be casted to a string.
        :returns boolean: ``True`` if the operation is successful.
        """
        return cls._instance.set(key, strval(value))

    @classmethod
    def setex(cls, key, value, second):
        """Set the key to the specified value, with an expiration time.

        :param string key: The Redis key.
        :param string value: The value to set.
        :param int second: The TTL in second.
        :returns boolean: ``True`` if the operation is successful.
        """
        return cls._instance.setex(key, second, strval(value))

    @classmethod
    def expire(cls, key, second):
        """Set the key to expire in specified second.

        :param string key: The key to set expire.
        :param int second: The number of seconds for the key to live.
        :returns boolean: True if the operation is successful.
        """
        return cls._instance.expire(key, second)

    @classmethod
    def rpush(cls, key, *values):
        """Add values to a Redis list from the end.

        The ``values`` argument is a variable-length array and can be
        specified as follows:

        .. code-block:: python

           redis.rpush('danqing', 'val1', 'val2')
           redis.rpush('danqing', 'val1')
           redis.rpush('danqing', 'val1', 'val2', 'val3')

        :param string key: The key of the list. If the list does not exist yet,
            it will be created.
        :param string... values: A list of values to insert. If any is not a
            string, it will be casted to a string.
        :returns int: The total number of elements in the list after the push.
        """
        return cls._instance.rpush(key, *strvals(*values))

    @classmethod
    def delete(cls, key):
        """Delete the key from Redis.

        :param string key: The key to delete.
        :returns int: The number of items deleted. If 1, the key is found and
            deleted. If 0, the key is not found and nothing is done.
        """
        return cls._instance.delete(key)

    @classmethod
    def hgetall(cls, key):
        """Get the hash table at the specified key.

        :param string key: The key to fetch.
        :returns dict: The hash table at the specified key. If no hash table
            found (i.e. key not found), an empty dictionary is returned.
        :raises redis.ResponseError: If ``key`` holds something other than a
            hash table.
        """
        return cls._instance.hgetall(key)

    @classmethod
    def hget(cls, key, hash_key):
        """Get the value for a hash key in the hash table at the specified key.

        :param string key: The key to fetch.
        :param string hash_key: The hash key to fetch value for in the hash
            table.
        :returns string: The value corresponding to the hash key in the hash
            table. If either ``key`` or ``hash_key`` is not found, ``None`` is
            returned.
        :raises redis.ResponseError: If ``key`` holds something other than a
            hash table.
        """
        return cls._instance.hget(key, hash_key)

    @classmethod
    def hset(cls, key, hash_key, hash_value):
        """Set the value for a hash key in the hash table at the specified key.

        :param string key: The key of the hash table. If the hash table does
            not exist yet, it will be created.
        :param string hash_key: The hash key to set value for in the hash
            table.
        :param string hash_value: The value to set to the key. If this is not
            a string, it will be casted to a string.
        :returns int: The number of new fields. If 1, a new field is added
            (``hash_key`` is new). If 0, ``hash_key`` already exists and its
            value is updated.
        :raises redis.ResponseError: If ``key`` holds something other than a
            hash table.
        """
        return cls._instance.hset(key, hash_key, strval(hash_value))

    @classmethod
    def hdelete(cls, key, *hash_keys):
        """Delete keys from a hash table.

        The ``hash_keys`` argument is a variable-length array and can be
        specified as follows:

        .. code-block:: python

           redis.hdelete('danqing', 'key1', 'key2')
           redis.hdelete('danqing', 'key2')
           redis.hdelete('danqing', 'key1', 'key2', 'key3')

        :param string key: The key of the hash table.
        :param string... hash_keys: A list of hash keys to delete from the hash
            table.
        :returns int: The number of keys actually deleted. If 3 hash keys are
            specified but only 1 is found (and deleted), 1 is returned.
        :raises redis.ResponseError: If ``key`` holds something other than a
            hash table.
        """
        return cls._instance.hdel(key, *hash_keys)

    @classmethod
    def lpeek(cls, key, count):
        """Peek the first count elements in the list without popping.

        :param string key: The key of the array.
        :param int count: The number of elements to peek.
        :returns list: The list of peeked elements.
        """
        return cls._instance.lrange(key, 0, count - 1)

    @classmethod
    def lpop(cls, key, count):
        """Pop the first count elements in the list.

        :param string key: The key of the array.
        :param int count: The number of elements to pop. If there are fewer
            than ``count`` elements, everything will be popped.
        :returns list: The list of popped elements.
        """
        pipe = cls._instance.pipeline()
        pipe.lrange(key, 0, count - 1)
        pipe.ltrim(key, count, - 1)
        result = pipe.execute()
        return result[0] if result[1] else []

    @classmethod
    def atomic_rw(cls, key, evaluator=lambda x: x):
        """Atomically read-write a Redis key.

        :param string key: The key to read/write.
        :param function evaluator: The evaluator function. It takes the
            existing value of the key, and should return the new value. If it
            returns None, no update is done and the operation is considered
            aborted. If not provided, the identity function is used.
        :returns *: The value returned by evaluator. If there's an
            atomicity violation, None is returned.
        :returns boolean: True if there's no atomicity error. That is, this
            value is True if write succeeded or the user aborted, and False if
            there's a atomicity violation (that prevented the write).
        """
        with cls._instance.pipeline() as pipe:
            try:
                pipe.watch(key)
                value = evaluator(pipe.get(key))
                if not value:
                    return None, True
                pipe.multi()
                pipe.set(key, strval(value))
                pipe.execute()
                return value, True
            except redis.WatchError:
                return None, False

    @classmethod
    def atomic_rw_hash(cls, key, hash_key, evaluator=lambda x: x):
        """Atomically read-write a Redis hash key.

        :param string key: The key of the hash to read/write.
        :param string hash_key: The hash key within the hash to read/write.
        :param function evaluator: The evaluator function. It takes the
            existing value of the key, and should return the new value. If it
            returns None, no update is done and the operation is considered
            aborted. If not provided, the identity function is used.
        :returns *: The value returned by evaluator. If there's an
            atomicity violation, None is returned.
        :returns boolean: True if there's no atomicity error. That is, this
            value is True if write succeeded or the user aborted, and False if
            there's a atomicity violation (that prevented the write).
        """
        with cls._instance.pipeline() as pipe:
            try:
                pipe.watch(key)
                value = evaluator(pipe.hget(key, hash_key))
                if not value:
                    return None, True
                pipe.multi()
                pipe.hset(key, hash_key, strval(value))
                pipe.execute()
                return value, True
            except redis.WatchError:
                return None, False



import unittest

from dq import string


class TestString(unittest.TestCase):

    def test_lower_no_punc(self):
        sin = 'Cornell is #1~~!!'
        sout = 'cornell is 1'
        assert string.lower_no_punc(sin) == sout

    def test_random_string(self):
        r1 = string.random_string(8)
        assert len(r1) == 8
        assert r1 != string.random_string(8)

    def test_valid_filename(self):
        assert string.valid_filename('a/b/ha ha.pdf') == 'abha_ha.pdf'

    def test_relative_path(self):
        assert not string.safe_relative_path(None)
        assert not string.safe_relative_path('')
        assert not string.safe_relative_path(123)
        assert string.safe_relative_path('hello')
        assert not string.safe_relative_path('/danqing')
        assert string.safe_relative_path('cornell/no1')
        assert string.safe_relative_path('cornell/no1..')
        assert string.safe_relative_path('cornell/no.1')
        assert string.safe_relative_path('cornell/no..1/forreal')
        assert not string.safe_relative_path('cornell/./no1')
        assert not string.safe_relative_path('cornell/../no1')
        assert not string.safe_relative_path('cornell/..//no1')
        assert not string.safe_relative_path('../danqing')
        assert not string.safe_relative_path('..//danqing')
        assert string.safe_relative_path('./danqing')
        assert not string.safe_relative_path('..')
        assert not string.safe_relative_path('/')
        assert not string.safe_relative_path('//')
        assert not string.safe_relative_path('/..')
        assert not string.safe_relative_path('//..')
        assert not string.safe_relative_path('../')
        assert not string.safe_relative_path('..//')
        assert string.safe_relative_path('cornell/.is/..no1')



import logging
import json
import pickle
from collections import namedtuple
from uuid import uuid4

import arrow
import redis
from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import as_declarative
from sqlalchemy_utils import ArrowType

from dq.config import Config
from dq.database import query_with_limit_offset, Session
from dq.logging import error

logger = logging.getLogger(__name__)

SERVICE_NAME = Config.get('meta.name')

# An attribute of a model. The order is the following:
# - name (string): The name of the attribute (the key)
# - allow_from (boolean): Whether the attribute can be inflated with incoming
#       dictionaries. Default is True.
# - allow_to (boolean): Whether the attribute can be exported to outgoing
#       dictionaries. Default is True.
# - serializer (function): The function that processes the attribute value
#       before converting to dictionary. Default is None (no-op).
# - deserializer (function): The function that processes the dictionary value
#       before converting to the object attribute. Default is None (no-op).
Attr = namedtuple('Attr', 'name allow_from allow_to serializer deserializer')
Attr.__new__.__defaults__ = (True, True, None, None)


def uuid4_string():
    """Return a UUID as string.

    The UUID type must NOT be used anywhere in the app - only use string.
    This function should be used whenever a new UUID is needed.

    :returns string: A UUID string.
    """
    return str(uuid4())


def arrow_in(value):
    """Convert an arrow-understandable time to an Arrow object.

    :param value: The arrow-understandable time or ``None``.
    :returns Arrow: The Arrow object, or ``None`` if the input is ``None``.
    """
    return arrow.get(value) if value else None


def arrow_out(value):
    """Convert an Arrow object to timestamp.

    :param Arrow value: The Arrow time or ``None``.
    :returns int: The timestamp, or ``None`` if the input is ``None``.
    """
    return value.timestamp if value else None


def boolean_mask(value):
    """Mask a value to boolean. This can be used for sensitive fields.

    :param value: Any value.
    :returns boolean: The input value casted to boolean.
    """
    return bool(value)


def enum_value(value):
    """Convert an enum to its value.

    :param Enum value: An enum.
    :returns string: The value of the enum.
    """
    return value.value if value else None


class DictMixin(object):
    """Dictionary mixin that provides converters to/from Python dicts.

    See comments of ``Attr`` above for defining fields to convert with this
    mixin.

    This class is inherited by ``IDBase`` and ``UUIDBase`` classes, so your
    class will not need to inherit directly from it.
    """

    __attrs__ = ()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.to_dict())

    def to_dict(self):
        """Convert an object to dictionary.

        :returns dict: The dictionary converted from the object.
        """
        out = {}
        for attr in self.__attrs__:
            if not attr.allow_to:
                continue
            name = attr.name
            out[name] = getattr(self, name)
            if attr.serializer:
                out[name] = attr.serializer(out[name])
        return out

    def to_json(self):
        """Convert an object to JSON string.

        :returns string: The JSON string representation of the object.
        """
        return json.dumps(self.to_dict())

    def inflate_to_dict(self, session=None):
        """Convert an object to dictionary, with relationships inflated.

        This method is the default implementation, which simply calls the
        ``to_dict`` method. Subclasses willing to inflate certain fields should
        override this method.

        :param Session session: The optional SQL session.
        :returns dict: The dictionary converted from the object.
        """
        return self.to_dict()

    @classmethod
    def from_dict(cls, data):
        """Convert a dictionary to an object.

        :param dict data: The data dictionary.
        :returns Base: The object converted from the dictionary.
        """
        entity = cls()
        for attr in entity.__attrs__:
            if not attr.allow_from:
                continue
            value = data.get(attr.name)
            if attr.deserializer:
                value = attr.deserializer(value)
            setattr(entity, attr.name, value)
        return entity


class TimeMixin(object):
    """Time fields common to all models.

    This class is inherited by ``IDBase`` and ``UUIDBase`` classes, so your
    class will not need to inherit directly from it.
    """

    created_at = Column(ArrowType, nullable=False, default=arrow.utcnow)
    updated_at = Column(ArrowType, nullable=False, default=arrow.utcnow,
                        onupdate=arrow.utcnow)
    deleted_at = Column(ArrowType)


class Cache(object):
    """Redis cache engine. This should not be used externally!

    This is a separate Redis connection from the one in the redis package. The
    purpose is twofold:

    1. The cache Redis does not allow response decoding, but the user Redis
       requires it.
    2. By separating the two, if a project does not have an explicit cache
       configuration, then cache can be effectively disabled globally.
    """

    _instance = None
    _attempted = False

    @classmethod
    def instance(cls):
        if cls._instance:
            return cls._instance
        if cls._attempted:
            return None
        cachecfg = Config.get('cache')
        if not cachecfg:
            cls._attempted = True
            return None
        try:
            i = redis.StrictRedis(**cachecfg)
            # This will attempt to connect to Redis and throw an error if the
            # connection is invalid.
            i.info()
            cls._instance = i
            return i
        except Exception:
            error(logger, 'Unable to connect to cache Redis', None)
            cls._attempted = True
            return None

    @classmethod
    def get(cls, key):
        """Get the value corresponding to the key in the cache.

        :param string key: The cache key.
        :returns bytes: The cache value in raw bytes if exists. If the key does
            not exist or if cache server cannot be reached, returns ``None``.
        """
        i = cls.instance()
        try:
            return i.get(key) if i else None
        except Exception as e:
            error(logger, 'Error querying cache', {'key': key, 'error': e})
            return None

    @classmethod
    def set(cls, key, value, ttl):
        i = cls.instance()
        if not i:
            return
        try:
            i.setex(key, ttl, pickle.dumps(value))
        except Exception as e:
            error(logger, 'Error setting cache', {'key': key, 'error': e})


class QueryMixin(object):
    """Query helper functions useful to all models.

    This class is inherited by ``IDBase`` and ``UUIDBase`` classes, so your
    class will not need to inherit directly from it.
    """

    @classmethod
    def cache_key(cls, column, value, contains_deleted=False,
                  contains_empty=False):
        key = '{}.cache.{}.{}.{}'.format(
            SERVICE_NAME, cls.__tablename__, column, value,
        )
        if contains_deleted:
            key += '.del'
        if contains_empty:
            key += '.empty'
        return key

    @classmethod
    def get_cache(cls, column, value, contains_deleted=False,
                  contains_empty=False):
        """Get the object from cache."""
        key = cls.cache_key(column, value, contains_deleted=contains_deleted,
                            contains_empty=contains_empty)
        cached = Cache.get(key)
        return pickle.loads(cached) if cached else None

    @classmethod
    def save_to_cache(cls, obj, column, value, ttl=3600,
                      contains_deleted=False, contains_empty=False):
        key = cls.cache_key(column, value, contains_deleted=contains_deleted,
                            contains_empty=contains_empty)
        Cache.set(key, obj, ttl)

    @classmethod
    def get_by(cls, column, value, contains_deleted=False,
               contains_empty=False, ttl=3600, fresh=False, session=None):
        """Get the object satisfying the query condition.

        :param string column: The name of the column to query by.
        :param string value: The value of the column to query for.
        :param boolean contains_deleted: Whether to contain deleted records.
            Default is ``False`` and only active records are returned.
        :param boolean contains_empty: Whether to contain empty records.
            Default is ``False`` and if value is ``None``, ``None`` will be
            returned.
        :param int ttl: The TTL of cache. If < 0, no cache will be used.
        :param boolean fresh: Whether to fetch the response fresh from the DB.
        :param Session session: Optional SQL session.
        :returns QueryMixin: The matching object. This method is designed for
            unique queries and in case of multiple matches, only the first one
            is returned.
        """
        if not contains_empty and value is None:
            return None

        if ttl > 0 and not fresh:
            cached = cls.get_cache(
                column, value, contains_deleted=contains_deleted,
                contains_empty=contains_empty,
            )
            if cached:
                return cached

        session = session or Session()
        args = {}
        args[column] = value
        if not contains_deleted:
            args['deleted_at'] = None
        obj = session.query(cls).filter_by(**args).first()
        if obj and ttl > 0:
            cls.save_to_cache(obj, column, value, ttl=ttl,
                              contains_deleted=contains_deleted,
                              contains_empty=contains_empty)
        return obj

    @classmethod
    def get_multi(cls, column, value, sort_column='updated_at', desc=True,
                  offset=0, limit=10, session=None):
        """Get multiple objects satisfying the query condition.

        :param string column: The name of the column to query by.
        :param string value: The value of the column to query for.
        :param string sort_column: The column to sort results against. Default
            is updated_at. Please note that created_at usually doesn't have
            index, so if it should be used as the sort column, an index needs
            to be added.
        :param bool desc: Whether to sort DESC. Default is ``True``.
        :param int offset: The query offset. Default is 0.
        :param int limit: The max number of results to return. Default is 10.
            If specified, the function will try to fetch one more result to
            indicate whether there are more results. If not specified or 0, all
            results will be returned.
        :param Session session: The optional SQL session to use.
        """
        session = session or Session()
        args = {'deleted_at': None}
        args[column] = value
        q = session.query(cls).filter_by(**args)
        if sort_column:
            attr = getattr(cls, sort_column)
            q = q.order_by(attr.desc()) if desc else q.order_by(attr.asc())
        return query_with_limit_offset(q, limit, offset).all()


class UserRelationMixin(QueryMixin):
    """Mixin that provides a common ``get_by_user`` function.

    This mixin should be used by classes that a user can have a 1-many
    relationship to, such as orders, shipments, and periods.
    """

    # The key used for sorting (DESC) when getting by user. This is most likely
    # a time field such as created_at or updated_at.
    sort_key = None

    @classmethod
    def get_by_user(cls, user_id, desc=True, offset=0, limit=10, session=None):
        """Get the related object of a user.

        :param string/int user_id: The ID (if int) or UUID (if string) of the
            user.
        :param bool desc: Whether to sort DESC. Default is ``True``.
        :param integer offset: The offset of the objects. If limit is 10 and
            offset is 5, the function will attempt to get the 6th to 15th
            most recent objects.
        :param integer limit: The max number of objects to get.
        :param Session session: The optional SQL session.
        :returns list<UserRelationMixin>: The list of objects matching the
            requirement.
        """
        column = 'user_id' if isinstance(user_id, int) else 'user_uuid'
        return cls.get_multi(column, user_id, cls.sort_key, desc=desc,
                             offset=offset, limit=limit, session=session)


@as_declarative()
class IDBase(QueryMixin, DictMixin, TimeMixin):
    """Base class for an ORM model with numeric ID as the primary key."""

    __attrs__ = (
        Attr('id', allow_to=False),
    )

    id = Column(Integer, primary_key=True)

    @classmethod
    def get(cls, id, contains_deleted=False, ttl=3600, fresh=False,
            session=None):
        """Get an object by its ID.

        :param int id: The ID of the object.
        :param boolean contains_deleted: Whether to contain deleted objects.
            Default is ``False``.
        :param int ttl: Cache TTL in seconds. If <= 0 no cache will be used.
            Default is 3600 (1 hour).
        :param boolean fresh: Whether to fetch a fresh copy even if cache
            exists. Default is ``False``.
        :param Session session: Optional SQL session to use.
        """
        return cls.get_by('id', id, contains_deleted=contains_deleted,
                          fresh=fresh, session=session)


@as_declarative()
class UUIDBase(QueryMixin, DictMixin, TimeMixin):
    """Base class for an ORM model with UUID as the primary key."""

    __attrs__ = (
        Attr('uuid', allow_to=False),
    )

    uuid = Column(String, primary_key=True, default=uuid4_string)

    @classmethod
    def get(cls, uuid, contains_deleted=False, ttl=3600, fresh=False,
            session=None):
        """Get an object by its UUID.

        :param string uuid: The UUID of the object.
        :param boolean contains_deleted: Whether to contain deleted objects.
            Default is ``False``.
        :param int ttl: Cache TTL in seconds. If <= 0 no cache will be used.
            Default is 3600 (1 hour).
        :param boolean fresh: Whether to fetch a fresh copy even if cache
            exists. Default is ``False``.
        :param Session session: Optional SQL session to use.
        """
        return cls.get_by('uuid', uuid, contains_deleted=contains_deleted,
                          fresh=fresh, session=session)



import random
import re

NAMESPACE = 'abcdefghijklmnopqrstuvwxyz1234567890'
PUNCS = '!"#$%&\'()*+,.:;<=>?@[\\]^_`{|}~'
TBL = dict.fromkeys(ord(i) for i in PUNCS)
SAFE_PATH = re.compile('^(?!(/|\.{2}))(?!.*/(\.*)(?:/.*|$)).*$')


def lower_no_punc(text):
    """Convert string to lowercase, stripping spaces and punctuations.

    :param string text: The text to clean up.
    :returns string: The cleaned up text.
    """
    # Strip whitespaces and convert to lowercase.
    return text.translate(TBL).strip().lower()


def random_string(length):
    """Get a random string of lowercase letters + numbers of the given length.

    :param int length: The desired length of the string.
    :returns string: A random string of the given length.
    """
    return ''.join(random.choice(NAMESPACE) for i in range(length))


def valid_filename(text):
    """Turn a string into a valid filename.

    :param string text: The string to convert.
    :returns string: A valid filename derived from the string.
    """
    text = text.strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', text)


def safe_relative_path(path):
    """Check whether a relative path is "safe".

    A path is safe if it is a relative path and doesn't contain . or ..'s.
    This prevents us from directory traversal attacks.

    Note that for simplicity, certain safe paths aren't allowed either, notably
    ... (e.g. .../path). Normal user should not really use such paths.

    :param string path: The path to check.
    :returns boolean: True if the path is safe, and False otherwise.
    """
    if not path or not isinstance(path, str):
        return False
    return bool(re.match(SAFE_PATH, path))



import ast
import json
import mock
import unittest
from contextlib import suppress
from uuid import uuid4

import arrow

from dq.database import commit_scope, save_to_database
from dq.orm import Cache
from tests.models import Table2, User, UserType


class TestORM(unittest.TestCase):

    def tearDown(self):
        with suppress(Exception), commit_scope() as session:
            session.query(Table2).delete()
            session.query(User).delete()
            if Cache._instance:
                Cache._instance.flushall()

    def test_string(self):
        t2 = Table2(id=1, key=1, key2=1, user_type=UserType.admin,
                    created_at=arrow.get('2017-10-21'))
        t2_str = str(t2)
        assert t2_str.startswith('<Table2 ')
        assert ast.literal_eval(t2_str[8:-1]) == {
            'id': 1,
            'user_uuid': None,
            'user_type': 'admin',
            'key2': True,
            'created_at': 1508544000,
        }

    def test_cache_key(self):
        assert Table2.cache_key('uuid', '123') == 'dqpy.cache.table2.uuid.123'
        k2 = Table2.cache_key('id', 123, contains_deleted=True,
                              contains_empty=True)
        assert k2 == 'dqpy.cache.table2.id.123.del.empty'

    def test_cached(self):
        t2 = Table2(id=999, user_uuid=str(uuid4()), key=1, key2=1,
                    user_type=UserType.regular)
        save_to_database(t2)

        key = 'dqpy.cache.table2.id.999'
        assert not Cache.get(key)

        t2 = Table2.get(999)
        assert t2.key == 1
        assert Cache.get(key)
        with commit_scope() as session:
            session.query(Table2).delete()
        t2 = Table2.get(999)
        assert t2.key == 1

    def test_cache_error(self):
        Cache._instance = 123
        assert not Cache.get('cornell')
        Cache.set('cornell', '#1', 123)
        Cache._instance = None

    @mock.patch('dq.config.Config.get')
    def test_cache_broken(self, mock_cfg):
        mock_cfg.return_value = {'port': 1234}
        Cache._instance = None
        Cache._attempted = None
        assert not Cache.instance()
        assert Cache._attempted
        assert not Cache.get('cornell')
        Cache.set('cornell', '#1', 123)
        Cache._attempted = None

    @mock.patch('dq.config.Config.get')
    def test_cache_none(self, mock_cfg):
        mock_cfg.return_value = None
        Cache._instance = None
        Cache._attempted = None
        assert not Cache.instance()
        assert Cache._attempted
        assert not Cache.get('cornell')
        Cache.set('cornell', '#1', 123)
        Cache._attempted = None

    def test_to_dict(self):
        uuid = str(uuid4())
        now = arrow.get()
        t2 = Table2(id=1, user_uuid=uuid, key=1, key2=1,
                    user_type=UserType.regular, created_at=now)

        t2_dict = t2.to_dict()
        t2_inflate = t2.inflate_to_dict()
        assert t2_inflate == t2_dict
        assert t2_dict == {
            'id': 1,
            'user_uuid': uuid,
            'user_type': 'regular',
            'key2': True,
            'created_at': now.timestamp,
        }

    def test_to_json(self):
        uuid = str(uuid4())
        now = arrow.get()
        t2 = Table2(id=1, user_uuid=uuid, key=1, key2=1,
                    user_type=UserType.regular, created_at=now)
        t2_json = t2.to_json()
        assert json.loads(t2_json) == {
            'id': 1,
            'user_uuid': uuid,
            'user_type': 'regular',
            'key2': True,
            'created_at': now.timestamp,
        }

    def test_from_dict(self):
        uuid = str(uuid4())
        now = arrow.get()
        t2_dict = {
            'id': 1,
            'user_uuid': uuid,
            'key': 1,
            'key2': 1,
            'created_at': now.timestamp,
        }
        t2 = Table2.from_dict(t2_dict)
        assert t2.created_at == arrow.get(now.timestamp)
        assert not t2.key2

    def test_get_by(self):
        uuid = str(uuid4())
        t2 = Table2(id=1, user_uuid=uuid)
        save_to_database(t2)

        t2 = Table2.get_by('user_uuid', uuid)
        assert t2.id == 1

    def test_get_by_empty(self):
        t2 = Table2(id=1, user_uuid=str(uuid4()))
        save_to_database(t2)
        assert not Table2.get_by('user_uuid', None)

    def test_get_multi(self):
        uuid = str(uuid4())
        t21 = Table2(id=1, user_uuid=uuid, key=1, key2=1)
        t22 = Table2(id=2, user_uuid=uuid, key=1, key2=2)
        t23 = Table2(id=3, user_uuid=uuid, key=1, key2=3)
        t24 = Table2(id=4, user_uuid=uuid, key=1, key2=4)
        t25 = Table2(id=5, user_uuid=uuid, key=1, key2=5)
        save_to_database([t21, t22, t23, t24, t25])

        results = Table2.get_multi('key', 1, 'key2', offset=2, limit=2)
        assert len(results) == 2
        assert results[0].key2 == 3
        assert results[1].key2 == 2

        results = Table2.get_multi('key', 1, 'key2', desc=False, limit=7)
        assert len(results) == 5
        assert results[0].key2 == 1

    def test_get_by_deleted(self):
        uuid = str(uuid4())
        t2 = Table2(id=1, user_uuid=uuid, deleted_at=arrow.get())
        save_to_database(t2)

        assert not Table2.get_by('user_uuid', uuid)

    def test_get_by_deleted_contains(self):
        uuid = str(uuid4())
        t2 = Table2(id=1, user_uuid=uuid)
        save_to_database(t2)

        t2 = Table2.get_by('user_uuid', uuid, contains_deleted=True)
        assert t2.id == 1

    def test_get_by_user(self):
        uuid = str(uuid4())
        t21 = Table2(id=1, user_uuid=uuid, key=1, key2=1,
                     created_at=arrow.get('2011-10-21'))
        t22 = Table2(id=2, user_uuid=uuid, key=1, key2=2,
                     created_at=arrow.get('2012-10-21'))
        t23 = Table2(id=3, user_uuid=uuid, key=1, key2=3,
                     created_at=arrow.get('2013-10-21'))
        t24 = Table2(id=4, user_uuid=uuid, key=1, key2=4,
                     created_at=arrow.get('2014-10-21'))
        t25 = Table2(id=5, user_uuid=uuid, key=1, key2=5,
                     created_at=arrow.get('2015-10-21'))
        save_to_database([t21, t22, t23, t24, t25])

        results = Table2.get_by_user(uuid, offset=2, limit=2)
        assert len(results) == 2
        assert results[0].key2 == 3
        assert results[1].key2 == 2



import unittest
from contextlib import suppress
from uuid import uuid4

from dq.database import (
    commit_scope, safe_save_to_database, save_to_database, Session)
from dq.errors import ModelError
from tests.models import Table2, User


class TestDatabase(unittest.TestCase):

    def tearDown(self):
        with suppress(Exception), commit_scope() as session:
            session.query(User).delete()
            session.query(Table2).delete()

    def test_replace_insert(self):
        uuid = str(uuid4())
        user = User(uuid=uuid, email='unittest@danqing.io')
        save_to_database(user)

        user = {'uuid': uuid, 'email': 'unittest2@danqing.io'}
        session = Session()
        session.execute(User.__table__.insert(mysql_replace_insert=True), user)
        session.commit()
        user = User.get(uuid)
        assert user.email == 'unittest2@danqing.io'

    def test_save_to_database_noop(self):
        save_to_database(None)

    def test_save_to_database_single(self):
        uuid = str(uuid4())
        user = User(uuid=uuid, email='unittest@danqing.io')
        save_to_database(user)

        user = User.get(uuid)
        assert user.email == 'unittest@danqing.io'

    def test_save_to_database_multi(self):
        uuid = str(uuid4())
        user = User(uuid=uuid, email='unittest@danqing.io')
        t2 = Table2(id=2, user_uuid=uuid)
        save_to_database([user, t2])

        user = User.get(uuid)
        assert user.email == 'unittest@danqing.io'
        t2 = Table2.get(2)
        assert t2.user_uuid == uuid

    def test_save_to_database_error(self):
        self.assertRaises(ModelError, save_to_database, User())

    def test_safe_save_to_database_noop(self):
        assert safe_save_to_database(None)

    def test_safe_save_to_database_single(self):
        uuid = str(uuid4())
        user = User(uuid=uuid, email='unittest@danqing.io')
        assert safe_save_to_database(user)

        user = User.get(uuid)
        assert user.email == 'unittest@danqing.io'

    def test_safe_save_to_database_multi(self):
        uuid = str(uuid4())
        user = User(uuid=uuid, email='unittest@danqing.io')
        t2 = Table2(id=1, user_uuid=uuid)
        assert safe_save_to_database([user, t2])

        user = User.get(uuid)
        assert user.email == 'unittest@danqing.io'
        t2 = Table2.get(1)
        assert t2.user_uuid == uuid

    def test_safe_save_to_database_error(self):
        assert not safe_save_to_database(User())



import logging
from functools import wraps

from dq.logging import error

logger = logging.getLogger(__name__)


def safe_invoke(noreturn=False):
    """Invoke a function safely.

    :param boolean noreturn: Whether the function being called has no return
        value itself. Default to False. If True and the function returns
        successfully, True will be returned. None will be returned upon failure
        no matter whether the function has return value.
    :returns func: The decorator function that provides safe invocation.
    """
    def safe_invoke_deco(func):
        """Decorator that captures an underlying error."""
        @wraps(func)
        def decorated_func(*args, **kwargs):
            try:
                retval = func(*args, **kwargs)
                return True if noreturn else retval
            except Exception as e:
                error(logger, 'Error invoking {}'.format(func.__qualname__), {
                    'function': func.__qualname__, 'module': func.__module__,
                    'args': args, 'kwargs': kwargs, 'error': e,
                })
                return None
        return decorated_func

    return safe_invoke_deco



import os

PRODUCTION = os.getenv('DQENV', 'local') == 'production'


def error(logger, message, payload=None):
    """Log an error message.

    :param logger logger: The logger to use.
    :param string message: The error message.
    :param dict payload: The payload to attach to the log.
    """
    _log(logger.error, message, payload, exc_info=True)


def warning(logger, message, payload=None):
    """Log a warning message.

    :param logger logger: The logger to use.
    :param string message: The warning message.
    :param dict payload: The payload to attach to the log.
    """
    _log(logger.warning, message, payload)


def info(logger, message, payload=None):
    """Log an info message.

    :param logger logger: The logger to use.
    :param string message: The info message.
    :param dict payload: The payload to attach to the log.
    """
    _log(logger.info, message, payload)


def debug(logger, message, payload=None):
    """Log a debug message. This does nothing in production enviornment.

    :param logger logger: The logger to use.
    :param string message: The debug message.
    :param dict payload: The payload to attach to the log.
    """
    if PRODUCTION:
        return
    _log(logger.debug, message, payload)


def _log(logger, message, payload, exc_info=False):
    """Log a message.

    This does two things:

    1. The payload (if present) is converted to a string and appended to the
       log message. Format string is used so that Sentry can group messages
       together correctly.
    2. The payload is attached again as a data dict. This doesn't show up in
       terminal/file, but Sentry uses it for searching.

    :param func logger: The logger with level to use (e.g. logger.error).
    :param string message: The log message.
    :param dict payload: The payload to attach to the log.
    :param boolean exc_info: Whether to log the stack trace if the log is for
        an exception.
    """
    if payload:
        message = message + '\n%s'
        pstr = _format(payload)
        logger(message, pstr, extra={'data': payload}, exc_info=exc_info)
    else:
        logger(message, exc_info=exc_info)


def _format(payload):
    """Format the payload (variables) as a string.

    When logging to the terminal or a file, only the string message is logged.
    This converts the payload into a string, which can then be appended to the
    log message itself.

    :param dict payload: The variable payload.
    :returns string: The formatted payload message.
    """
    message = ''
    for key in payload.keys():
        message += '\n`{}`: {}'.format(key, payload[key])
    return message



import unittest

from tests.entities import SimpleEntity


class TestEntity(unittest.TestCase):

    def test_string(self):
        e = SimpleEntity({'name': 'danqing'})
        assert str(e) == "<SimpleEntity {'name': 'danqing'}>"



import os
import unittest
from contextlib import suppress

from dq import fsutil


class TestFSUtil(unittest.TestCase):

    def tearDown(self):
        with suppress(Exception):
            fsutil.rmrf('test-files')

    def test_safe_mkdirp(self):
        fsutil.mkdirp('test-files/1/2/3')
        assert os.path.isdir('test-files/1/2/3')

    def test_safe_rmrf(self):
        fsutil.mkdirp('test-files/1/2')
        with open('test-files/1/2/3.txt', 'w') as f:
            f.write('hello')
        assert os.path.exists('test-files/1/2/3.txt')
        fsutil.rmrf('test-files/1/2/3.txt')
        assert not os.path.exists('test-files/1/2/3.txt')
        with open('test-files/1/2/3.txt', 'w') as f:
            f.write('hello')
        fsutil.rmrf('test-files')
        assert not os.path.exists('test-files')
        fsutil.rmrf('none')

    def test_traverse_error(self):
        self.assertRaises(FileNotFoundError, fsutil.traverse, 'none', None)
        self.assertRaises(
            NotADirectoryError, fsutil.traverse, 'README.md', None,
        )

    def test_traverse(self):
        fsutil.mkdirp('test-files/1/2')
        with open('test-files/1/2/3.txt', 'w') as f:
            f.write('hello')

        tree = {}

        def callback(path, isdir):
            assert path not in tree.keys()
            tree[path] = isdir

        fsutil.traverse('test-files', callback)
        assert tree == {'1': True, '1/2': True, '1/2/3.txt': False}

    def test_fileinfo(self):
        fsutil.mkdirp('test-files')
        with open('test-files/a.txt', 'w') as f:
            f.write('hello')
        assert fsutil.fileinfo('test-files/a.txt') == ('text/plain', 5, None)

    def test_safe_fileinfo(self):
        assert fsutil.safe_fileinfo(None) == (None, 0, None)



import unittest

from dq import util


class TestUtil(unittest.TestCase):

    def test_safe_cast(self):
        assert util.safe_cast('1', int) == 1
        assert util.safe_cast('meow', int, 2) == 2



import mock
import unittest

from pymysql.err import OperationalError
from sqlalchemy import exc

from dq.errors import RecoverableError
from dq import retry


class TestRetry(unittest.TestCase):

    def test_recoverable_error(self):
        e = RecoverableError('recoverable')
        assert retry.recoverable_error(e)

    @mock.patch('dq.retry.Session.rollback')
    def test_sql_or_recoverable_error_sql(self, mock_rollback):
        e = OperationalError(2013, 'meow')
        assert retry.sql_or_recoverable_error(e)
        mock_rollback.assert_called()

    @mock.patch('dq.retry.Session.rollback')
    def test_sql_or_recoverable_error_sqlalchemy(self, mock_rollback):
        e = exc.TimeoutError()
        assert retry.sql_or_recoverable_error(e)
        mock_rollback.assert_called()

    @mock.patch('dq.retry.Session.rollback')
    def test_sql_or_recoverable_error_sqlalchemy_not(self, mock_rollback):
        e = exc.NoSuchTableError()
        assert not retry.sql_or_recoverable_error(e)
        mock_rollback.assert_called()

    def test_sql_or_recoverable_error_recoverable(self):
        e = RecoverableError('recoverable')
        assert retry.sql_or_recoverable_error(e)



import mock
import unittest
from logging import Logger

from dq import logging


class TestLogging(unittest.TestCase):

    def test_error(self):
        logger = Logger(__name__)
        logger.error = mock.MagicMock()
        data = {'a': 'b'}
        logging.error(logger, 'hi', data)
        logger.error.assert_called_with(
            'hi\n%s', '\n`a`: b', exc_info=True, extra={'data': data},
        )

    def test_error_no_payload(self):
        logger = Logger(__name__)
        logger.error = mock.MagicMock()
        logging.error(logger, 'hi')
        logger.error.assert_called_with('hi', exc_info=True)

    def test_warning(self):
        logger = Logger(__name__)
        logger.warning = mock.MagicMock()
        logging.warning(logger, 'hi', {'a': 'b'})
        logger.warning.assert_called_with(
            'hi\n%s', '\n`a`: b', exc_info=False, extra={'data': {'a': 'b'}},
        )

    def test_info(self):
        logger = Logger(__name__)
        logger.info = mock.MagicMock()
        logging.info(logger, 'hi', {'a': 'b'})
        logger.info.assert_called_with(
            'hi\n%s', '\n`a`: b', exc_info=False, extra={'data': {'a': 'b'}},
        )

    def test_debug_production(self):
        production = logging.PRODUCTION
        logging.PRODUCTION = True
        logger = Logger(__name__)
        logger.debug = mock.MagicMock()
        logging.debug(logger, 'hi', {'a': 'b'})
        logger.debug.assert_not_called()
        logging.PRODUCTION = production

    def test_debug(self):
        logger = Logger(__name__)
        logger.debug = mock.MagicMock()
        logging.debug(logger, 'hi', {'a': 'b'})
        logger.debug.assert_called_with(
            'hi\n%s', '\n`a`: b', exc_info=False, extra={'data': {'a': 'b'}},
        )



class DQError(Exception):
    """Base exception.

    Clients that wish to catch all exceptions should catch this one.
    The ``status_code`` will be set as the HTTP status code if a view handler
    triggers an error. If there is need to set the exact type of error, use
    ``error_code``.

    4xx errors are due to the user, and as such the message will be passed
    directly back to user. 5xx errors will have the message replaced by the
    generic 'DQ Error' before sending back to user.
    """

    def __init__(self, message='', status_code=None, error_code=None):
        Exception.__init__(self)
        self.message = message
        self.status_code = status_code or 500
        self.error_code = error_code

    def __repr__(self):
        return '<{} - status: {}, message: {}>'.format(
            self.__class__.__name__, self.status_code, self.message
        )

    def __str__(self):
        code = str(self.status_code) + ' - ' if self.status_code else ''
        message = self.message if self.message else 'Unspecified DQ Error'
        return code + message

    def to_dict(self):
        message = self.message if self.status_code < 500 else 'DQ Error'
        return {'error': message}


class RecoverableError(DQError):
    """Exception raised when a recoverable error occurs.

    This may be due to network error, timeout, etc. and is meant to be retried.
    If retrying gives no luck, this is considered a 500 error.
    """

    def __init__(self, message):
        DQError.__init__(self, message=message, status_code=500)


class ModelError(DQError):
    """Exception raised when something is wrong with the models."""

    def __init__(self, message):
        DQError.__init__(self, message=message, status_code=500)


class IntegrationError(DQError):
    """Exception raised when there is a service integration error."""

    def __init__(self, message):
        DQError.__init__(self, message=message, status_code=500)


class InvalidInputError(DQError):
    """Exception raised when the user input is not valid."""

    def __init__(self, message='Invalid input'):
        DQError.__init__(self, message=message, status_code=400)


class NotFoundError(DQError):
    """Exception raised when something is not found."""

    def __init__(self, message='Not found'):
        DQError.__init__(self, message=message, status_code=404)


class UnauthorizedError(DQError):
    """Exception raised when the user is not authenticated for an action."""

    def __init__(self, message='Unauthorized'):
        DQError.__init__(self, message=message, status_code=401)


class ForbiddenError(DQError):
    """Exception raised when the user does not have the required permission."""

    def __init__(self, message='Forbidden'):
        DQError.__init__(self, message=message, status_code=403)



import os

import toml


class Config(object):
    """Configuration helper."""

    _instance = None

    @classmethod
    def init(cls):
        """Initialize the configuration.

        If the config file contains a dictionary keyed by ``env``, all its
        content will be set as environment variables.

        This function should not be called externally. The ``Config.get``
        function lazily initializes the config object automatically.
        """
        dqenv = os.getenv('DQENV') or 'local'
        with open('config/{}.toml'.format(dqenv)) as conffile:
            cls._instance = toml.loads(conffile.read())
            env = cls._instance.get('env')
            if isinstance(env, dict):
                for key, value in env.items():
                    os.environ[key] = value

    @classmethod
    def get(cls, key):
        """Get the config value specified by the provided key.

        TOML config files should be stored in the config folder, and the
        environment (i.e. which file to choose) should be specified by the
        ``DQENV`` environment variable.

        :param string key: The key to retrieve the value for. Nested key should
            have its components separated by dot, such as ``owner.email``.
        :returns: The value of the key if it exists, and ``None`` otherwise.
        """
        if cls._instance is None:
            cls.init()

        parts = key.split('.')
        cfg = cls._instance
        for part in parts:
            cfg = cfg.get(part)
            if cfg is None:
                return None

        return cfg



import subprocess
import unittest

from sqlalchemy_utils import database_exists


class TestDatabase(unittest.TestCase):

    def test_create_drop(self):
        db = 'mysql+pymysql://root@127.0.0.1:3306/dqpy_dbadmin?charset=utf8mb4'
        subprocess.Popen(
            'dbadmin create dbadmin_sql.url',
            shell=True,
            stdout=subprocess.PIPE,
        ).stdout.read()
        assert database_exists(db)

        subprocess.Popen(
            'dbadmin drop dbadmin_sql.url',
            shell=True,
            stdout=subprocess.PIPE,
        ).stdout.read()
        assert not database_exists(db)

    def test_create_error(self):
        error = subprocess.Popen(
            'dbadmin create wrong', shell=True, stdout=subprocess.PIPE,
        ).stdout.read()
        assert 'Unable to create database' in str(error)

    def test_drop_error(self):
        error = subprocess.Popen(
            'dbadmin drop wrong', shell=True, stdout=subprocess.PIPE,
        ).stdout.read()
        assert 'Unable to drop database' in str(error)



from pymysql.err import OperationalError
from sqlalchemy import exc

from dq.database import Session
from dq.errors import RecoverableError


def recoverable_error(exception):
    """Use as ``@retry(retry_on_exception=recoverable_error)``."""
    return isinstance(exception, RecoverableError)


def sql_or_recoverable_error(exception):
    """Retries recoverable error and SQL errors.

    This function rolls back the current session if the exception is a SQL one.
    Use as ``@retry(retry_on_exception=sql_or_recoverable_error)``.
    """
    if isinstance(exception, OperationalError):
        if exception.args[0] in (2006, 2013, 2014, 2045, 2055):
            Session.rollback()
            return True
    if isinstance(exception, exc.SQLAlchemyError):
        Session.rollback()
        return isinstance(exception, (
            exc.DisconnectionError, exc.StatementError, exc.TimeoutError,
            exc.DBAPIError,
        ))
    return recoverable_error(exception)



#!/usr/bin/env python

from setuptools import find_packages, setup

__version__ = '1.3.1'

requires = [
    'arrow==0.12.1',
    'emails==0.5.15',
    'PyMySQL==0.8.1',
    'redis==2.10.6',
    'schematics==2.0.1',
    'SQLAlchemy-Utils==0.33.3',
    'SQLAlchemy==1.2.8',
    'toml==0.9.4',
]

setup(
    name='dqpy',
    version=__version__,
    description='Danqing\'s shared Python library',
    author='Danqing Liu',
    author_email='code@danqing.io',
    url='https://github.com/danqing/dqpy',
    packages=find_packages(),
    install_requires=requires,
    entry_points={
        'console_scripts': ['dbadmin=dq.dbadmin:main'],
    },
    zip_safe=True,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ),
)



def safe_cast(value, to_type, default=None):
    """Cast a value to another type safely.

    :param value: The original value.
    :param type to_type: The destination type.
    :param default: The default value.
    """
    try:
        return to_type(value)
    except (ValueError, TypeError):
        return default



from schematics.models import Model


class Entity(Model):
    """Base entity class that provides shared utilities.

    Currently it only allows printing a schematics entity. More utilities
    may be added in the future. In any case, all entities should inherit from
    this class.
    """

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.to_primitive())



import unittest

from dq.funcutil import safe_invoke


@safe_invoke()
def with_return(name, action=None):
    if action != 'raise':
        return action
    raise Exception


@safe_invoke(noreturn=True)
def no_return(name, action=None):
    if action == 'raise':
        raise Exception
    if action == 'force':
        return 'force'


class TestFuncUtil(unittest.TestCase):

    def test_safe_invoke_success_return(self):
        assert with_return('danqing', action='hi') == 'hi'
        assert with_return('danqing') is None

    def test_safe_invoke_success_noreturn(self):
        assert no_return('danqing', action='hi') is True
        assert no_return('danqing', action='force') is True

    def test_safe_invoke_error_return(self):
        assert with_return('danqing', action='raise') is None

    def test_safe_invoke_error_noreturn(self):
        assert no_return('danqing', action='raise') is None



import sys

from sqlalchemy import create_engine
from sqlalchemy_utils import create_database, database_exists, drop_database

from dq.config import Config

ACTION_CREATE = 'create'
ACTION_DROP = 'drop'


def create(url_key):
    try:
        engine = create_engine(Config.get(url_key))
        if not database_exists(engine.url):
            create_database(engine.url)
    except Exception as e:
        print('Unable to create database: %s' % e)


def drop(url_key):
    try:
        engine = create_engine(Config.get(url_key))
        drop_database(engine.url)
    except Exception as e:
        print('Unable to drop database: %s' % e)


def main():
    action = sys.argv[1]
    url_key = sys.argv[2] if len(sys.argv) > 2 else 'sql.url'
    if action == ACTION_CREATE:
        create(url_key)
    elif action == ACTION_DROP:
        drop(url_key)



from enum import Enum

from sqlalchemy import Column, Integer, String
from sqlalchemy_utils import ChoiceType

from dq import orm


class UserType(Enum):

    regular = 'regular'
    admin = 'admin'


class User(orm.UUIDBase):

    __tablename__ = 'users'
    __attrs__ = (
        orm.Attr('email'),
    )

    email = Column(String)


class Table2(orm.IDBase, orm.UserRelationMixin):

    __tablename__ = 'table2'
    __attrs__ = (
        orm.Attr('id'),
        orm.Attr('user_uuid'),
        orm.Attr('user_type', serializer=orm.enum_value),
        orm.Attr('key', allow_to=False),
        orm.Attr('key2', allow_from=False, serializer=orm.boolean_mask),
        orm.Attr('created_at', serializer=orm.arrow_out,
                 deserializer=orm.arrow_in),
    )

    sort_key = 'created_at'

    user_uuid = Column(String)
    user_type = Column(ChoiceType(UserType, impl=String()))
    key = Column(Integer)
    key2 = Column(Integer)
